import use_psyco, section
import os

use_psyco.null()

_SRILM_LM_FILE_PATH = "/tmp/"
_SRILM_LM_TEMP_FILE_PATH = "/tmp/lm.tmp"
_SRILM_PATH = "/afs/csail.mit.edu/group/rbg/tools/bin/language_model/srilm/"

class LanguageModel:

    def __init__( self, all_sections, training_sections ):

        self.unigram_scores = {}
        self.unigram_backoffs = {}
        self.bigram_scores = {}
        
        order = 2
        
        self.lm_file = _SRILM_LM_FILE_PATH + "lm_arpa_" + str(order) + ".lmfile"
        
        # Write text to file
        train_text_file = _SRILM_LM_TEMP_FILE_PATH
        f = open(train_text_file, 'w')
        for sec in all_sections:
            for sent in sec.body:
                f.write( sent.text.lower() + "\n" )
        for sec in training_sections:
            f.write( sec.title.text.lower() + "\n" )
        f.close()
            
        # Train language model with SRILM
        print "* Training...",
        lp_pipe = os.popen( _SRILM_PATH + "ngram-count -order " + str(order) + " -text " + train_text_file + " -lm " + self.lm_file)
        lp_pipe.close()
        print "Done."

        self.__read_arpa_output(self.lm_file)
        
    def __read_arpa_output(self, arpa_file):

        f = open(arpa_file, 'r')
        arpa_file_contents = f.read()
        f.close()

        lines = arpa_file_contents.split("\n")

        level = 0
        for line in lines:

            line = line.strip()

            # Find start of unigrams
            if level == 0:

                if "\\1-grams:" in line:
                    level = 1
                    continue
                else:
                    continue

            # Read unigrams, find start of bigrams
            elif level == 1:

                if "\\2-grams:" in line:

                    level = 2
                    continue

                elif line == "":
                    continue
                else:

                    line_parts = line.split("\t")

                    if len(line_parts) == 2:
                        
                        [score, unigram] = line_parts
                        self.unigram_scores[unigram] = float(score)
                        
                    elif len(line_parts) == 3:

                        [score, unigram, backoff] = line_parts
                        self.unigram_backoffs[unigram] = float(backoff)
                        self.unigram_scores[unigram] = float(score)

                    else:

                        raise Exception("Unknown line split: "+line)
                        
            # Read bigrams
            elif level == 2:
            
                if line == "" or line == "\\end\\":
                    continue

                else:

                    [score, bigram] = line.split("\t")
                    [first_word, second_word] = bigram.split(" ")

                    self.bigram_scores[(first_word, second_word)] = float(score)
            else:
                
                raise Exception("Unknown level "+str(level))
                
    def get(self, first_word, second_word):

        return ( self.bigram_scores.get((first_word, second_word), \
                                        self.unigram_scores.get(second_word, 0.0)+self.unigram_backoffs.get(first_word, 0.0)) )



class TagLanguageModel:

    def __init__( self, training_sections ):

        self.unigram_scores = {}
        self.unigram_backoffs = {}
        self.bigram_scores = {}
        
        order = 2
        
        self.lm_file = _SRILM_LM_FILE_PATH + "lm_arpa_" + str(order) + ".lmfile"
        
        # Write text to file
        train_text_file = _SRILM_LM_TEMP_FILE_PATH
        f = open(train_text_file, 'w')
        for sec in training_sections:
            #f.write( ' '.join( [ x.pos_tag for x in sec.title.words ] ) + "\n" )
            f.write( ' '.join( [ x.pos_tag for x in sec.title.words ] ) + ' ' + section.END_WORD.pos_tag + "\n" )
            #f.write( section.START_WORD.pos_tag + ' ' + ' '.join( [ x.pos_tag for x in sec.title.words ] ) + ' ' + section.END_WORD.pos_tag + "\n" )
        f.close()
            
        # Train language model with SRILM
        print "* Training...",
        lp_pipe = os.popen( _SRILM_PATH + "ngram-count -order " + str(order) + " -text " + train_text_file + " -lm " + self.lm_file)
        lp_pipe.close()
        print "Done."

        self.__read_arpa_output(self.lm_file)
        
    def __read_arpa_output(self, arpa_file):

        f = open(arpa_file, 'r')
        arpa_file_contents = f.read()
        f.close()

        lines = arpa_file_contents.split("\n")

        level = 0
        for line in lines:

            line = line.strip()

            # Find start of unigrams
            if level == 0:

                if "\\1-grams:" in line:
                    level = 1
                    continue
                else:
                    continue

            # Read unigrams, find start of bigrams
            elif level == 1:

                if "\\2-grams:" in line:

                    level = 2
                    continue

                elif line == "":
                    continue
                else:

                    line_parts = line.split("\t")

                    if len(line_parts) == 2:
                        
                        [score, unigram] = line_parts
                        self.unigram_scores[unigram] = float(score)
                        
                    elif len(line_parts) == 3:

                        [score, unigram, backoff] = line_parts
                        self.unigram_backoffs[unigram] = float(backoff)
                        self.unigram_scores[unigram] = float(score)

                    else:

                        raise Exception("Unknown line split: "+line)
                        
            # Read bigrams
            elif level == 2:
            
                if line == "" or line == "\\end\\":
                    continue

                else:

                    [score, bigram] = line.split("\t")
                    [first_word, second_word] = bigram.split(" ")

                    self.bigram_scores[(first_word, second_word)] = float(score)
            else:
                
                raise Exception("Unknown level "+str(level))
                
    def get(self, first_word, second_word):

        return ( self.bigram_scores.get((first_word, second_word), \
                                        self.unigram_scores.get(second_word, 0.0)+self.unigram_backoffs.get(first_word, 0.0)) )

