import config, generative_lm, decoder, use_psyco, utils, reader, section
import math, sys, os, getopt

use_psyco.null()

_MAXENT_DATA_PATH = "/tmp/mxtrain.data"

_ZHANGLE_MAXENT_MODEL_PATH = "/tmp/maxent.model"
_ZHANGLE_MAXENT_TEST_DIR = "/tmp/"

def mapfv( word, sec, root_sec ):
    fv_map = {}

    all_sections = root_sec.get_descendants ()
    section_count = len (all_sections)
    
    # Lexicalized
    fv_map[ word.word ] = 1

    # Feature: first occurrence in section by sentence, without normalization
    #feat_first_occurrence_by_sentence = sec.get_first_occurrence_by_sentence( word.stem )
    #fv_map[ "fobs" ] = feat_first_occurrence_by_sentence
    
    # Feature: first occurrence in section by word       
    sec.first_occurrence_by_word[word.stem] = sec.first_occurrence_by_word.get (word.stem, \
                                                                                1.0 * sec.get_first_occurrence_by_word( word.stem ) / sec.get_total_unigram_occurrences())
    s = sec.first_occurrence_by_word[word.stem]
    if s <= 0.25:
        fv_map[ "fobw1" ] = 1
    elif s <= 0.50:
        fv_map[ "fobw2" ] = 1
    elif s <= 0.75:
        fv_map[ "fobw3" ] = 1
    elif s <= 1.0:
        fv_map[ "fobw4" ] = 1
    
    # Feature TF IDF
    #feat_tf = 1.0 * sec.get_unigram_occurrence( word.stem, True ) / sec.get_total_unigram_occurrences()
    #feat_df = 1.0 * root_sec.get_num_descendants_with_unigram( word.stem, True) / section_count
    #fv_map[ "tf-idf" ] = feat_tf * ( -1.0 / math.log( feat_df ) )
    
    return fv_map

def render_zhangle ( fv_map, in_title ):
    fv_string = (in_title and "t" or "f") + " "
    fv_string += ' '.join( [ str(feature) + ":" + str(value) for feature, value in fv_map.items() ] )

    return fv_string

def train_selection_zhangle( root_sec, train_start, train_end ):
    print "* Training selection model..."

    # Generate training file
    print " * Generating training file..."
    f = open( _MAXENT_DATA_PATH , 'w' )
    for i, sec in enumerate( root_sec.get_descendants()[ train_start : train_end + 1] ):

        if len( sec.title.words ) == 0: continue

        title_words = set( [ x.word for x in sec.title.words ] )

        print "    " +str(i) + sec.title.text
        
        for word in sec.unique_word_objs:
            f.write( render_zhangle( mapfv( word, sec, root_sec ) , (word.word in title_words) ) + "\n" )
            
    f.close()

    # Train
    print " * Running maximum entropy..."
    p = os.popen( config.MAXENT_ZHANGLE_EXEC_PATH + " " + _MAXENT_DATA_PATH + " -b -m " + _ZHANGLE_MAXENT_MODEL_PATH + " -i 14 -v" )
    print p.read()
    p.close()

def get_selection_scores_zhangle( sec, root_sec ):
    print "* Calculating selection scores..."
    selection_words = []

    test_path = os.tempnam (_ZHANGLE_MAXENT_TEST_DIR, 'mxtest.')
    output_path = _ZHANGLE_MAXENT_TEST_DIR + "maxent.out"

    # Generate testing file
    print " * Generating test file..."
    f = open( test_path , 'w' )
    for word in sec.unique_word_objs:
        f.write( render_zhangle( mapfv( word, sec, root_sec ) , False ) + "\n" )        
        selection_words.append( word.word )
    f.close()

    # Test
    print " * Running maximum entropy..."
    p = os.popen( config.MAXENT_ZHANGLE_EXEC_PATH + " -p -b -m " + _ZHANGLE_MAXENT_MODEL_PATH \
                  + " --detail --output=" + output_path + " " + test_path)
    p.close()
    os.unlink (test_path)
    
    # Scrape output
    print " * Scraping output..."

    selection_scores = {}
    
    f = open(output_path, 'r')
    lines =  f.readlines()
    f.close()
    os.unlink (output_path)
    
    for word, line in zip( selection_words, lines ):        
        [class1, prob1, class2, prob2] = line.strip().split("\t")
        if class1 == "t": selection_scores[ word ] = float(prob1)
        if class2 == "t": selection_scores[ word ] = float(prob2)

    return selection_scores


def decode( root_section, beam_size, train_start, train_end, test_start, test_end, outfile ):

    lm = build_lexical_language_model( root_section, train_start, train_end )
    taglm = build_tag_language_model( root_section )

    print "* Decoding..."

    f = open( outfile , 'w')
    f.write( '== sub tree =====================\n')
    
    for sec in root_section.get_descendants()[ test_start : test_end + 1 ]:

        print " * Generating title for Section : " + sec.title.text + "..."
        
        selection_scores = get_selection_scores_zhangle( sec, root_section )
        length = len( sec.title.words )

        if length == 0: continue
        
        bq = decoder.BeamQueue( beam_size )
        #bq.add_initial( [ section.START_WORD ], 0 )
        bq.add_initial( [ ], 0 ) 

        for i in range( length ):

            new_bq = decoder.BeamQueue( beam_size )

            for (path_words, path_score) in bq.items():
                for word in sec.unique_word_objs:

                    # Skip rendundant stems
                    if (word.stem in map (lambda x : x.stem, path_words)):
                        continue

                    new_path_score = path_score + selection_scores[ word.word ]
                    if len( path_words ) > 0:
                        new_path_score += lm.get( path_words[-1].word, word.word )
                        new_path_score += taglm.get( path_words[-1].pos_tag, word.pos_tag )

                    if i == length - 1:
                        new_path_score += lm.get( word.word, section.END_WORD.word )
                        new_path_score += taglm.get( word.pos_tag, section.END_WORD.pos_tag )
                    
                    new_bq.add( path_words, word, new_path_score )

            bq = new_bq

        print
        print "TOC : " + ' '.join( [ x.word for x in bq.get_best()[:] ] ) 
        print "ACT : " + ' '.join( [ x.word for x in sec.title.words ] )
        print
        sys.stdout.flush()
        
        f.write( "TOC : " + ' '.join( [ x.word for x in bq.get_best()[:] ] ) + "\n" )
        f.write( "ACT : " + ' '.join( [ x.word for x in sec.title.words ] ) + "\n\n" )
        
    f.close()
        
def build_lexical_language_model( root_section, train_start, train_end ):
    print "* Building lexical language model..."
    all_sections = root_section.get_descendants()
    return generative_lm.LanguageModel ( all_sections, all_sections[ train_start: train_end ] )

def build_tag_language_model( root_section ):
    print "* Building tag language model..."
    all_sections = root_section.get_descendants()    
    return generative_lm.TagLanguageModel ( all_sections )

def load_dataset():
    utils.p( "* Reading corpus..." )
    try:
        utils.p("* Loading from pickle...")
        return utils.load_section()
    except:
        utils.p("* Loading from XML...")
        return reader.read_corpus()

def main():
    
    try:
        opts, args = getopt.getopt (sys.argv[1:], "",\
                                    [ "mode=", "dataset=", "beam_size=", "train_start=", "train_end=", "test_start=", "test_end=", "outfile=" ])

    except getopt.GetoptError:
        print "Invalid arguments supplied"
        sys.exit(2)

    def opt(name):
        key = "--"+name
        for option, val in opts:
            if option == key:
                return val
        return None

    config.DATASET_NAME = opt("dataset")
    root_section = load_dataset()

    if( opt("mode") == "decode" ):
        decode( root_section, int( opt("beam_size") ), int( opt("train_start") ), int( opt("train_end") ), int( opt("test_start") ), int( opt("test_end") ), opt("outfile") )

    elif( opt("mode") == "train" ):
        train_selection_zhangle( root_section, int( opt("train_start") ), int( opt("train_end") ) )

if __name__ == '__main__':
    sys.exit( main( ) )





####### YKLEE #################################
'''

_YKLEE_MAXENT_CLASSIFIER_PATH = "/afs/csail.mit.edu/group/rbg/tools/bin/classifiers/maxent_yklee/me2"
_YKLEE_MAXENT_MODEL_PATH = "/tmp/"
_YKLEE_MAXENT_TEST_DIR = "/tmp/"

def train_selection_yklee( root_sec, train_start, train_end ):
    print "* Training selection model..."

    # Generate training file
    print " * Generating training file..."
    f = open( _MAXENT_DATA_PATH , 'w' )
    for i, sec in enumerate( root_sec.get_descendants()[ train_start : train_end + 1] ):
        title_words = set( [ x.word for x in sec.title.words ] )

        print "    " +str(i) + sec.title.text
        
        for word in sec.unique_word_objs:
            f.write( render_fv_yklee( mapfv( word, sec, root_sec ) , (word.word in title_words) ) + "\n" )
            
    f.close()

    # Train
    print " * Running maximum entropy..."
    p = os.popen( _YKLEE_MAXENT_CLASSIFIER_PATH + " -l -m " + _YKLEE_MAXENT_MODEL_PATH + " < " + _MAXENT_DATA_PATH )
    p.close()

def get_selection_scores_yklee( sec, root_sec ):
    print "* Calculating selection scores..."
    selection_words = []

    test_path = os.tempnam (_YKLEE_MAXENT_TEST_DIR, 'mxtest.')

    # Generate testing file
    print " * Generating test file..."
    f = open( test_path , 'w' )
    for word in sec.unique_word_objs:
        f.write( render_fv_yklee( mapfv( word, sec, root_sec ) , False ) + "\n" )        
        selection_words.append( word.word )
    f.close()

    # Test
    print " * Running maximum entropy..."
    p = os.popen( _YKLEE_MAXENT_CLASSIFIER_PATH + " -m " + _YKLEE_MAXENT_MODEL_PATH + " < " + test_path )
    lines = p.readlines()
    p.close()

    os.unlink( test_path )

    # Scrape output
    print " * Scraping output..."
    selection_scores = {}
    for word, line in zip( selection_words, lines ):
        parts = line.split(" ")
        selection_scores[ word ] = math.log10( float( parts[ parts.index("+title") + 1 ] ) )
    
    return selection_scores

def render_fv_yklee( fv_map, in_title ):
    fv = in_title and "+title " or "-title "

    fv += ' '.join( [ str(feature) + " " + str(value) for feature, value in fv_map ] )
    return fv
'''
